# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from authlib.flask.oauth2 import AuthorizationServer, ResourceProtector
from authlib.flask.oauth2.sqla import (
    create_query_client_func,
    create_save_token_func,
    create_revocation_endpoint,
    create_bearer_token_validator,
)
from authlib.specs.rfc6749 import grants
from authlib.specs.rfc6749.errors import InvalidRequestError
from werkzeug.security import gen_salt
from .models import db, User, OAuth2Client, OAuth2AuthorizationCode, OAuth2Token


class AuthorizationCodeGrant(grants.AuthorizationCodeGrant):
    def create_authorization_code(self, client, user, request):
        code = gen_salt(48)
        item = OAuth2AuthorizationCode(
            code=code,
            client_id=client.client_id,
            redirect_uri=request.redirect_uri,
            scope=request.scope,
            user_id=user.id,
        )
        db.session.add(item)
        db.session.commit()
        return code

    def parse_authorization_code(self, code, client):
        item = OAuth2AuthorizationCode.query.filter_by(
            code=code, client_id=client.client_id).first()
        if item and not item.is_expired():
            return item

    def delete_authorization_code(self, authorization_code):
        db.session.delete(authorization_code)
        db.session.commit()

    def authenticate_user(self, authorization_code):
        return User.query.get(authorization_code.user_id)


class PasswordGrant(grants.ResourceOwnerPasswordCredentialsGrant):
    def authenticate_user(self, username, password):
        user = User.query.filter_by(username=username).first()
        if user is None:
            raise InvalidRequestError("User Not Found")
        if user.check_password(password):
            return user


class RefreshTokenGrant(grants.RefreshTokenGrant):
    def authenticate_refresh_token(self, refresh_token):
        item = OAuth2Token.query.filter_by(refresh_token=refresh_token).first()
        if item and not item.is_refresh_token_expired():
            return item

    def authenticate_user(self, credential):
        return User.query.get(credential.user_id)


query_client = create_query_client_func(db.session, OAuth2Client)
save_token = create_save_token_func(db.session, OAuth2Token)
# Authlib 0.8 has a bug in init_app
authorization = AuthorizationServer(
    query_client=query_client,
    save_token=save_token,
)
require_oauth = ResourceProtector()


def config_oauth(app):
    authorization.init_app(app)

    # support all grants
    authorization.register_grant(grants.ImplicitGrant)
    authorization.register_grant(grants.ClientCredentialsGrant)
    authorization.register_grant(AuthorizationCodeGrant)
    authorization.register_grant(PasswordGrant)
    authorization.register_grant(RefreshTokenGrant)

    # support revocation
    revocation_cls = create_revocation_endpoint(db.session, OAuth2Token)
    authorization.register_endpoint(revocation_cls)

    # protect resource
    bearer_cls = create_bearer_token_validator(db.session, OAuth2Token)
    require_oauth.register_token_validator(bearer_cls())



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from datetime import timedelta
from .constants import NMOSAUTH_DIR, PRIVKEY_PATH, DATABASE_NAME

pkg = ''
if __package__ is not None:
    pkg = __package__ + '.'


class BaseConfig(object):
    DEBUG = True
    TESTING = False
    SECRET_KEY = os.urandom(16)
    PERMANENT_SESSION_LIFETIME = timedelta(minutes=5)
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    SQLALCHEMY_DATABASE_URI = 'sqlite:///{}/{}.sqlite'.format(NMOSAUTH_DIR, DATABASE_NAME)
    BASIC_AUTH_REALM = "NMOS Auth Server Login Required"
    OAUTH2_ACCESS_TOKEN_GENERATOR = pkg + 'token_generator.gen_token'
    OAUTH2_REFRESH_TOKEN_GENERATOR = True
    OAUTH2_JWT_ENABLED = True
    OAUTH2_JWT_ISS = 'http://rd.bbc.co.uk/x-nmos/auth/v1.0/'
    OAUTH2_JWT_ALG = 'RS256'
    OAUTH2_JWT_EXP = 60
    OAUTH2_JWT_KEY_PATH = PRIVKEY_PATH


class TestConfig(BaseConfig):
    DEBUG = True
    TESTING = True
    SQLALCHEMY_DATABASE_URI = 'sqlite:///:memory:'
    OAUTH2_JWT_KEY_PATH = None


class ProductionConfig(BaseConfig):
    DEBUG = False



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from authlib.common.errors import AuthlibBaseError, AuthlibHTTPError
from flask import jsonify, render_template


def register_handlers(app):

    @app.errorhandler(AuthlibHTTPError)
    def authlib_http_handler(error):
        e = error()
        status = e[0]
        body = e[1]
        headers = e[2]
        return (jsonify(body), status, headers)

    @app.errorhandler(AuthlibBaseError)
    def authlib_base_handler(error):
        return jsonify(error=str(error)), 400

    @app.errorhandler(401)
    def page_unauthorised(e):
        code = 401
        return render_template('error.html', code=code, message="Unauthorised"), code

    @app.errorhandler(404)
    def page_not_found(e):
        code = 404
        return render_template('error.html', code=code, message="Not Found"), code

    @app.errorhandler(403)
    def page_forbiddon(e):
        code = 403
        return render_template('error.html', code=code, message="Forbiddon"), code



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, absolute_import
from gevent import monkey
monkey.patch_all()

import gevent  # noqa E402
import time  # noqa E402
import signal  # noqa E402
from socket import gethostname  # noqa E402
from os import getpid  # noqa E402

from nmoscommon.httpserver import HttpServer  # noqa E402
from nmoscommon.mdns import MDNSEngine  # noqa E402
from nmoscommon.logger import Logger  # noqa E402
from nmoscommon.nmoscommonconfig import config as _config  # noqa E402
from .security_api import SecurityAPI  # noqa E402


PORT = 4999
HOSTNAME = gethostname().split(".", 1)[0]
API_VERSIONS = ["v1.0"]

DNS_SD_HTTP_PORT = 80
DNS_SD_HTTPS_PORT = 443
DNS_SD_NAME = 'auth_' + str(HOSTNAME) + "_" + str(getpid())
DNS_SD_TYPE = '_nmos-auth._tcp'


class SecurityService:
    def __init__(self, logger=None):
        self.config = _config
        self.logger = Logger("nmosauth", logger)
        self.running = False
        self.httpServer = None
        self.mdns = MDNSEngine()

    def start(self):
        if self.running:
            gevent.signal(signal.SIGINT, self.sig_handler)
            gevent.signal(signal.SIGTERM, self.sig_handler)

        self.mdns.start()

        priority = self.config["priority"]
        if not str(priority).isdigit():
            priority = 0

        if self.config["https_mode"] != "enabled":
            self.mdns.register(DNS_SD_NAME + "_http", DNS_SD_TYPE, DNS_SD_HTTP_PORT,
                               {"pri": priority,
                                "api_ver": ",".join(API_VERSIONS),
                                "api_proto": "http"})
        if self.config["https_mode"] != "disabled":
            self.mdns.register(DNS_SD_NAME + "_https", DNS_SD_TYPE, DNS_SD_HTTPS_PORT,
                               {"pri": priority,
                                "api_ver": ",".join(API_VERSIONS),
                                "api_proto": "https"})

        self.httpServer = HttpServer(
            SecurityAPI, PORT, '0.0.0.0', api_args=[self.logger, self.config, 'ProductionConfig', None]
        )
        self.httpServer.start()
        while not self.httpServer.started.is_set():
            self.logger.writeInfo('Waiting for httpserver to start...')
            self.httpServer.started.wait()

        if self.httpServer.failed is not None:
            raise self.httpServer.failed

        self.logger.writeInfo("Running on port: {}".format(self.httpServer.port))

    def run(self):
        self.running = True
        self.start()
        while self.running:
            time.sleep(1)
        self._cleanup()

    def _cleanup(self):
        if self.mdns:
            try:
                self.mdns.stop()
                self.mdns.close()
                self.logger.writeInfo("mDNS stopped gracefully")
            except Exception as e:
                self.logger.writeWarning("Could not stop mDNS gracefully: {}".format(e))

        self.httpServer.stop()
        self.logger.writeInfo("Stopped Http Server")

    def sig_handler(self):
        self.logger.writeInfo('Pressed ctrl+c')
        self.stop()

    def stop(self):
        self.running = False


if __name__ == '__main__':
    Service = SecurityService()
    Service.run()



#!/usr/bin/python

from OpenSSL import crypto
import os

from nmosauth.auth_server.constants import CERT_PATH, PRIVKEY_PATH


def create_self_signed_cert():

    # create a key pair
    k = crypto.PKey()
    k.generate_key(crypto.TYPE_RSA, 2048)

    # create a self-signed cert
    cert = crypto.X509()
    cert.get_subject().C = "UK"
    cert.get_subject().ST = "London"
    cert.get_subject().L = "London"
    cert.get_subject().O = "Dummy Company Ltd"  # noqa: E741
    cert.get_subject().OU = "Dummy Company Ltd"
    cert.get_subject().CN = "www.example.com"
    cert.set_serial_number(1000)
    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(10 * 365 * 24 * 60 * 60)
    cert.set_issuer(cert.get_subject())
    cert.set_pubkey(k)
    cert.sign(k, 'sha256')

    # Write Cert and Private Key to File
    open(CERT_PATH, "wt").write(
        (crypto.dump_certificate(crypto.FILETYPE_PEM, cert).decode('utf-8'))
    )
    open(PRIVKEY_PATH, "wt").write(
        crypto.dump_privatekey(crypto.FILETYPE_PEM, k).decode('utf-8')
    )

    # Change permissions to Read-Only for security
    os.chmod(CERT_PATH, 0o400)
    os.chmod(PRIVKEY_PATH, 0o400)


create_self_signed_cert()



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from flask import request, session, send_from_directory
from flask import render_template, redirect, url_for, jsonify
from werkzeug.security import gen_salt
from jinja2 import FileSystemLoader, ChoiceLoader
from authlib.specs.rfc6749 import OAuth2Error
from nmoscommon.webapi import WebAPI, route
from nmoscommon.auth.nmos_auth import RequiresAuth

from .models import db, User, OAuth2Client
from .oauth2 import authorization
from .app import config_app
from .basic_auth import basicAuth
from .db_utils import getUser, removeClient, addUser, addAccessRights
from .constants import CERT_PATH

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))

APINAMESPACE = "x-nmos"
APINAME = "auth"
APIVERSION = "v1.0"

AUTH_API_ROOT = '/{}/{}'.format(APINAMESPACE, APINAME)
AUTH_VERSION_ROOT = '{}/{}/'.format(AUTH_API_ROOT, APIVERSION)


class SecurityAPI(WebAPI):
    def __init__(self, logger, nmosConfig, confClass, extraConfig=None):
        super(SecurityAPI, self).__init__()
        self._config = nmosConfig
        self.logger = logger
        self.add_templates_folder()
        config_app(self.app, confClass=confClass, config=extraConfig)  # OAuth and DB config

    # Add html templates folder to list of Jinja loaders
    def add_templates_folder(self):
        my_loader = ChoiceLoader([
            self.app.jinja_loader,
            FileSystemLoader(SCRIPT_DIR + '/static'),
            FileSystemLoader(SCRIPT_DIR + '/templates')
        ])
        self.app.jinja_loader = my_loader

    def current_user(self):
        if 'id' in session:
            uid = session['id']
            return User.query.get(uid)
        return None

    # Custom function to serve CSS and Javascript files
    @route(AUTH_VERSION_ROOT + 'static/<filename>', auto_json=False)
    def style(self, filename):
        return send_from_directory(SCRIPT_DIR + '/static', filename)

    @route('/')
    def index(self):
        return (200, [APINAMESPACE + "/"])

    @route('/' + APINAMESPACE + "/")
    def namespaceindex(self):
        return (200, [APINAME + "/"])

    @route(AUTH_API_ROOT + '/')
    def nameindex(self):
        return (200, [APIVERSION + "/"])

    @route(AUTH_VERSION_ROOT)
    def versionindex(self):
        obj = ["home/", "signup/", "register_client/", "fetch_token/", "revoke/", "authorize/", "token/", "certs/"]
        return (200, obj)

    @route(AUTH_VERSION_ROOT + 'test/', auto_json=True)
    @RequiresAuth(condition=True)
    def test(self):
        return (200, "Hello World")

    @route(AUTH_VERSION_ROOT + 'home/', methods=['GET', 'POST'], auto_json=False)
    def home(self):
        if request.method == 'POST':
            username = request.form.get('username')
            password = request.form.get('password')
            if not username or not password:
                message = "Please Fill In Both Username and Password."
                return render_template('home.html', user=None, clients=None, message=message)
            user = User.query.filter_by(username=username).first()
            if not user:
                message = "That username is not recognised. Please signup."
                return render_template('home.html', user=None, clients=None, message=message)
            if user.password == password:
                session['id'] = user.id
                return redirect(url_for('_home'))
            else:
                message = "Invalid Password. Try Again."
                return render_template('home.html', user=None, clients=None, message=message)
        user = self.current_user()
        if user:
            clients = OAuth2Client.query.filter_by(user_id=user.id).all()
        else:
            clients = []
        return render_template('home.html', user=user, clients=clients, message="")

    @route(AUTH_VERSION_ROOT + 'signup', methods=['POST'], auto_json=False)
    def signup_post(self):
        username = request.form.get('username', None)
        password = request.form.get('password', None)
        if not username or not password:
            return redirect(url_for('_signup_get'))
        user = addUser(username, password)

        is04 = request.form.get('is04', None)
        is05 = request.form.get('is05', None)
        addAccessRights(user, is04, is05)

        session['id'] = user.id
        return redirect(url_for('_home'))

    @route(AUTH_VERSION_ROOT + 'signup/', methods=['GET'], auto_json=False)
    def signup_get(self):
        return render_template('signup.html')

    @route(AUTH_VERSION_ROOT + 'register_client', methods=['POST'], auto_json=False)
    @basicAuth.required
    def create_client_post(self):
        user = self.current_user()
        if not user and request.authorization:
            user = getUser(request.authorization.username)
        client = OAuth2Client(**request.form.to_dict(flat=True))
        client.user_id = user.id
        client.client_id = gen_salt(24)
        if client.token_endpoint_auth_method == 'none':
            client.client_secret = ''
        else:
            client.client_secret = gen_salt(48)
        db.session.add(client)
        db.session.commit()
        if 'id' in session:
            return redirect(url_for('_home'))
        else:
            client_info = client.client_info.copy()
            client_info.update(client.client_metadata)
            return jsonify(client_info), 201

    @route(AUTH_VERSION_ROOT + 'register_client/', methods=['GET'], auto_json=False)
    def create_client_get(self):
        user = self.current_user()
        if not user:
            return redirect(url_for('_home'))
        return render_template('create_client.html')

    @route(AUTH_VERSION_ROOT + 'delete_client/<client_id>', auto_json=False)
    def delete_client(self, client_id):
        removeClient(client_id)
        return redirect(url_for('_home'))

    @route(AUTH_VERSION_ROOT + 'fetch_token/', auto_json=False)
    def fetch_token(self):
        user = self.current_user()
        if not user:
            return redirect(url_for('_home'))
        # TODO - drop-down select box
        client = OAuth2Client.query.filter_by(user_id=user.id).first()
        return render_template('fetch_token.html', client=client)

    @route(AUTH_VERSION_ROOT + 'authorize', methods=['POST'], auto_json=False)
    def authorization_post(self):
        user = self.current_user()
        if not user and request.authorization:
            user = getUser(request.authorization.username)
        if not user and 'username' in request.form:
            username = request.form.get('username')
            user = User.query.filter_by(username=username).first()
        if request.form['confirm']:
            grant_user = user
        else:
            grant_user = None
        return authorization.create_authorization_response(grant_user=grant_user)

    @route(AUTH_VERSION_ROOT + 'authorize/', methods=['GET'], auto_json=False)
    def authorization_get(self):
        user = self.current_user()
        if not user and request.authorization:
            user = getUser(request.authorization.username)
        try:
            grant = authorization.validate_consent_request(end_user=user, request=request)
        except OAuth2Error as error:
            return error.error
        return render_template('authorize.html', user=user, grant=grant)

    @route(AUTH_VERSION_ROOT + 'token', methods=['POST'], auto_json=False)
    def issue_token_post(self):
        return authorization.create_token_response()

    @route(AUTH_VERSION_ROOT + 'revoke', methods=['POST'], auto_json=False)
    def revoke_token_post(self):
        return authorization.create_endpoint_response('revocation')

    # route for certificate with public key
    @route(AUTH_VERSION_ROOT + 'certs/', methods=['GET'], auto_json=True)
    def get_cert(self):
        try:
            with open(CERT_PATH, 'r') as myfile:
                cert = myfile.read()
            return (200, [cert])
        except OSError as e:
            self.logger.writeError("Error: {}\nFile at {} doesn't exist".format(e, CERT_PATH))
            raise

    @route(AUTH_VERSION_ROOT + 'logout/', auto_json=False)
    def logout(self):
        try:
            del session['id']
        except Exception as e:
            self.logger.writeDebug("Error: {}. Couldn't delete session ID".format(str(e)))
        return redirect(url_for('_home'))



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from flask_cors import CORS
from .models import db
from .oauth2 import config_oauth
from .handlers import register_handlers
from .basic_auth import basicAuth


def config_app(app, confClass='BaseConfig', config=None):

    # load configuration
    settings = 'settings'
    if __package__ is not None:
        settings = __package__ + '.settings'
    app.config.from_object(settings + '.' + confClass)

    # load environment configuration
    if 'WEBSITE_CONF' in os.environ:
        app.config.from_envvar('WEBSITE_CONF')

    # load app specified configuration
    if config is not None:
        if isinstance(config, dict):
            app.config.update(config)
        elif config.endswith('.py'):
            app.config.from_pyfile(config)

    setup_app(app)
    return app


def setup_app(app):
    db.init_app(app)
    basicAuth.init_app(app)
    config_oauth(app)
    register_handlers(app)
    CORS(app)

    # create all db tables
    with app.app_context():
        db.create_all()



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from authlib.specs.rfc7519 import jwt
import datetime
from .oauth2 import authorization
from .models import AccessRights
from .constants import NMOSAUTH_DIR, PRIVKEY_FILE


class TokenGenerator():

    def __init__(self):
        pass

    def get_access_rights(self, user, scope):
        if scope == "is04":
            user_access = AccessRights.query.filter_by(user_id=user.id).first()
            access = user_access.is04
        elif scope == "is05":
            user_access = AccessRights.query.filter_by(user_id=user.id).first()
            access = user_access.is05
        else:
            access = None
        return access

    def get_audience(self, user, scope, access):
        if access is None:
            return None
        if scope == "is04":
            if access == "write":
                audience = "IS-04 Write Access".split(" ")
            if access == "read":
                audience = "IS-04 Read Access".split(" ")
        elif scope == "is05":
            if access == "write":
                audience = "IS-05 Write Access".split(" ")
            if access == "read":
                audience = "IS-04 Read Access".split(" ")
        else:
            audience = None
        return audience

    def get_scope(self, user, client, scope):

        if scope in ["is04", "IS04", "is-04", "IS-04"]:
            new_scope = "is-04"
        elif scope in ["is05", "IS05", "is-05", "IS-05"]:
            new_scope = "is-05"
        else:
            new_scope = None
        return new_scope

    def gen_access_token(self, client, grant_type, user, scope):

        config = authorization.config
        current_time = datetime.datetime.utcnow()
        access = self.get_access_rights(user, scope)
        audience = self.get_audience(user, scope, access)
        new_scope = self.get_scope(user, client, scope)

        header = {
            "alg": config["jwt_alg"],
            "typ": "JWT"
        }
        payload = {
            'iat': current_time,
            'exp': current_time + datetime.timedelta(seconds=config['jwt_exp']),
            'nbf': current_time,
            'iss': config['jwt_iss'],
            'sub': user.username,
            'scope': new_scope,
            'aud': audience,
            'x-nmos-api': {
                'name': new_scope,
                'access': access
            }
        }

        try:
            key = config['jwt_key']
        except Exception as e:
            print("Error: " + e)
            abs_key_path = os.path.join(NMOSAUTH_DIR, PRIVKEY_FILE)
            with open(abs_key_path, 'r') as myfile:
                key = myfile.read()

        return jwt.encode(
            header,
            payload,
            key
        ).decode('utf-8')


# Needed for access_token path in Flask config
gen_token = TokenGenerator().gen_access_token



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import wraps
from flask import Response, request, current_app, render_template

from .models import User


class BasicAuthorization():
    """
    A Flask extension for adding HTTP basic access authentication to the
    application.
    """

    def __init__(self, app=None):
        if app is not None:
            self.app = app
            self.init_app(app)
        else:
            self.app = None

    def init_app(self, app):
        app.config.setdefault('BASIC_AUTH_REALM', '')

    def check_credentials(self, username, password):
        try:
            user = User.query.filter_by(username=username).first()
            return username == user.username and password == user.password
        except Exception:
            return False

    def authenticate(self):
        auth = request.authorization
        return (
            auth and auth.type == 'basic' and self.check_credentials(auth.username, auth.password)
        )

    def challenge(self):
        realm = current_app.config['BASIC_AUTH_REALM']
        return Response(
            status=401,
            headers={'WWW-Authenticate': 'Basic realm="{}"'.format(realm)},
            response=render_template('error.html', code=401, message="Unauthorised")
        )

    def required(self, view_func):
        @wraps(view_func)
        def wrapper(*args, **kwargs):
            if self.authenticate():
                return view_func(*args, **kwargs)
            else:
                return self.challenge()
        return wrapper


basicAuth = BasicAuthorization()



#!/usr/bin/python
#
# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import absolute_import
import unittest
import mock


from nmoscommon.logger import Logger
from nmosauth.auth_server.db_utils import drop_all
from nmosauth.auth_server.security_api import SecurityAPI
from nmosauth.auth_server.security_api import User
from nmos_auth_data import BEARER_TOKEN, TEST_PRIV_KEY
from base64 import b64encode

VERSION_ROOT = '/x-nmos/auth/v1.0'


class TestNmosAuth(unittest.TestCase):

    def setUp(self):
        self.api = SecurityAPI(logger=Logger("testing"), nmosConfig=None,
                               extraConfig={"OAUTH2_JWT_KEY": TEST_PRIV_KEY},
                               confClass='TestConfig')
        self.app = self.api.app
        self.client = self.app.test_client()

        self.mockUser = self.createMockUser("steve", "password")
        patcher = mock.patch("nmosauth.auth_server.basic_auth.User")
        self.mockBasicUser = patcher.start()
        self.mockBasicUser.query.filter_by.return_value.first.return_value = self.mockUser
        self.addCleanup(patcher.stop)

    def tearDown(self):
        with self.app.app_context():
            drop_all()

    def createMockUser(self, username, password):
        tempUser = User()
        tempUser.username = username
        tempUser.password = password
        return tempUser

    def auth_headers(self, user):
        auth_string = "{}:{}".format(user.username, user.password).encode('utf-8')
        headers = {
            'Authorization': b'Basic ' + b64encode(auth_string)
        }
        return headers

    def testGetInitialRoutes(self):

        with self.client as client:
            rv = client.get(VERSION_ROOT + '/', follow_redirects=True)
            self.assertEqual(rv.status_code, 200)
            rv = client.get(VERSION_ROOT + '/fetch_token/')
            self.assertEqual(rv.status_code, 302)
            rv = client.get(VERSION_ROOT + '/logout/')
            self.assertEqual(rv.status_code, 302)

    def testBasicAuthRoutes(self):

        headers = self.auth_headers(self.mockUser)
        with self.client as client:
            rv = client.get(VERSION_ROOT + '/register_client/')
            self.assertEqual(rv.status_code, 302)

            rv = client.post(VERSION_ROOT + '/register_client')
            self.assertEqual(rv.status_code, 401)

            wrongUser = self.createMockUser("bob", "pass")
            headers = self.auth_headers(wrongUser)
            rv = client.post(VERSION_ROOT + '/register_client', headers=headers)
            self.assertEqual(rv.status_code, 401)

    @mock.patch("nmosauth.auth_server.security_api.render_template")
    @mock.patch("nmosauth.auth_server.security_api.User")
    def testHome(self, mockUser, mockTemplate):

        mockTemplate.return_value = "test"

        with self.client.post(VERSION_ROOT + '/home/', data=dict(username="", password="")):
            mockTemplate.assert_called_with(
                'home.html', clients=None, message='Please Fill In Both Username and Password.', user=None)
        with self.client.post(VERSION_ROOT + '/home/', data=dict(username="steve", password="wrongpassword")):
            mockTemplate.assert_called_with(
                'home.html', clients=None, message='Invalid Password. Try Again.', user=None)
        with self.client.post(VERSION_ROOT + '/home/', data=dict(username="steve", password="password")) as rv:
            self.assertEqual(rv.status_code, 200)
        mockUser.query.filter_by.return_value.first.return_value = None
        with self.client.post(VERSION_ROOT + '/home/', data=dict(username="steve", password="password")):
            mockTemplate.assert_called_with(
                'home.html', clients=None, message='That username is not recognised. Please signup.', user=None)

    @mock.patch("nmosauth.auth_server.security_api.authorization")
    def testTokenEndpoint(self, mockAuthServer):
        mockAuthServer.create_endpoint_response.return_value = BEARER_TOKEN


if __name__ == '__main__':
    unittest.main()



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Constants for nmos-auth defining filenames and directory locations"""

import os

NMOSAUTH_DIR = '/var/nmosauth'
CERT_FILE = 'certificate.pem'
CERT_PATH = os.path.join(NMOSAUTH_DIR, CERT_FILE)
PRIVKEY_FILE = 'privkey.pem'
PRIVKEY_PATH = os.path.join(NMOSAUTH_DIR, PRIVKEY_FILE)
CERT_ENDPOINT = '/certs'
CERT_KEY = 'default'
DATABASE_NAME = 'auth_db'



#!/usr/bin/python
#
# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import os
from setuptools import setup
import subprocess
from setuptools.command.develop import develop
from setuptools.command.install import install

from nmosauth.auth_server.constants import NMOSAUTH_DIR

GEN_CERT_FILE = 'gen_cert.py'
GEN_CERT_PATH = os.path.join(NMOSAUTH_DIR, GEN_CERT_FILE)


# Basic metadata
name = "nmos-auth"
version = "1.0.4"
description = "OAuth2 Server Implementation"
url = 'https://github.com/bbc/nmos-auth-server'
author = 'Danny Meloy'
author_email = 'danny.meloy@bbc.co.uk'
license = 'Apache2'
long_description = "OAuth2 Server Implementation to produce JWTs for API Access"


def gen_certs():
    try:
        subprocess.call([GEN_CERT_PATH])
    except Exception as e:
        print('Error: {}. Failed to generate certificates.'.format(str(e)))
        print('Please run "{}" in {}'.format(GEN_CERT_FILE, NMOSAUTH_DIR))
        pass


class PostDevelopCommand(develop):
    """Post-installation for development mode."""
    def run(self):
        develop.run(self)
        gen_certs()


class PostInstallCommand(install):
    """Post-installation for installation mode."""
    def run(self):
        install.run(self)
        gen_certs()


def is_package(path):
    return (
        os.path.isdir(path) and os.path.isfile(os.path.join(path, '__init__.py'))
    )


def find_packages(path, base=""):
    """ Find all packages in path """
    packages = {}
    for item in os.listdir(path):
        dir = os.path.join(path, item)
        if is_package(dir):
            if base:
                module_name = "%(base)s.%(item)s" % vars()
            else:
                module_name = item
            packages[module_name] = dir
            packages.update(find_packages(dir, module_name))
    return packages


packages = find_packages(".")
package_names = packages.keys()

# This is where you list packages which are required
packages_required = [
    "six",
    "flask",
    "sqlalchemy",
    "flask-sqlalchemy",
    "flask-cors",
    "requests",
    "gevent",
    "nmoscommon",
    "pyopenssl",
    "authlib>=0.11"
]

deps_required = []

setup(
    name=name,
    version=version,
    description=description,
    url=url,
    author=author,
    author_email=author_email,
    license=license,
    packages=package_names,
    package_dir=packages,
    install_requires=packages_required,
    include_package_data=True,
    scripts=[],
    package_data={
        'nmosauth': ['auth_server/templates/*', 'auth_server/static/*']
    },
    data_files=[
        ('/usr/bin', ['bin/nmosauth']),
        (NMOSAUTH_DIR, ['nmosauth/certs/{}'.format(GEN_CERT_FILE)])
    ],
    long_description=long_description,
    cmdclass={
        'develop': PostDevelopCommand,
        'install': PostInstallCommand
    }
)



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from flask_sqlalchemy import SQLAlchemy
from authlib.flask.oauth2.sqla import (
    OAuth2ClientMixin,
    OAuth2AuthorizationCodeMixin,
    OAuth2TokenMixin,
)
from authlib.specs.rfc6749.errors import OAuth2Error

__all__ = ['db', 'User', 'OAuth2Client',
           'OAuth2AuthorizationCode', 'OAuth2Token', 'AccessRights']

db = SQLAlchemy()


class User(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(40), unique=True)
    password = db.Column(db.String(20))

    def __str__(self):
        output = ''
        for c in self.__table__.columns:
            output += '{}: {},  '.format(c.name, getattr(self, c.name))
        return output

    def get_user_id(self):
        return self.id

    def check_password(self, password):
        return password == self.password


class OAuth2Client(db.Model, OAuth2ClientMixin):
    __tablename__ = 'oauth2_client'

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(
        db.Integer, db.ForeignKey('user.id', ondelete='CASCADE'))
    user = db.relationship('User')

    def check_requested_scopes(self, scopes):  # Just for testing purposes
        if len(set(scopes)) != 1:
            raise OAuth2Error("Must list a single scope", None, 400)
        return super(OAuth2Client, self).check_requested_scopes(scopes)


class OAuth2AuthorizationCode(db.Model, OAuth2AuthorizationCodeMixin):
    __tablename__ = 'oauth2_code'

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(
        db.Integer, db.ForeignKey('user.id', ondelete='CASCADE'))
    user = db.relationship('User')


class OAuth2Token(db.Model, OAuth2TokenMixin):
    __tablename__ = 'oauth2_token'

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(
        db.Integer, db.ForeignKey('user.id', ondelete='CASCADE'))
    user = db.relationship('User')

    def is_refresh_token_expired(self):
        expires_at = self.issued_at + self.expires_in * 2
        return expires_at < time.time()


class AccessRights(db.Model):
    __tablename__ = 'access_rights'

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(
        db.Integer, db.ForeignKey('user.id', ondelete='CASCADE'))
    user = db.relationship('User')
    is04 = db.Column(db.String(25))
    is05 = db.Column(db.String(25))

    def __str__(self):
        output = ''
        for c in self.__table__.columns:
            output += '{}: {},  '.format(c.name, getattr(self, c.name))
        return output



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .models import db, User, OAuth2Client, AccessRights

# -------------------- INIT ------------------------- #

# engine = db.engine
# inspector = db.inspect(engine)
# metadata = db.metadata(engine)


# Drop all tables and re-create
def clean_start():
    drop_all()
    create_all()

# -------------------- CREATE ------------------------- #


def create_all():
    db.create_all()
    # db.session.commit()


def add(entry):
    db.session.add(entry)
    db.session.commit()


def addAccessRights(user, IS04Access, IS05Access):
    access = AccessRights(user_id=user.id, is04=IS04Access, is05=IS05Access)
    add(access)
    return access


def addUser(username, password):
    user = User(username=username, password=password)
    add(user)
    return user

# -------------------- READ ------------------------- #
# from nmosauth.auth_server.db_utils import *


def printTables():
    for table in db.metadata.tables:
        print("Table: {}".format(table))


def printTable(table):
    return [i for i in table.query.all()]


def printTableString(table):
    return [str(i) for i in table.query.all()]


def printField(table, field):
    s = "[i." + field + " for i in table.query.all()]"
    return eval(s)


def getUser(user):
    if isinstance(user, int):
        entry = User.query.get_or_404(user)
    elif isinstance(user, str):
        entry = User.query.filter_by(username=user).first_or_404()
    return entry


def printForeign(table, key, val):  # Key is usually user_id
    s = "db.session.query(" + table + ").filter(" + table + "." + key + "==" + str(val) + ").all()"
    return eval(s)

# -------------------- UPDATE ------------------------- #


# DANGEROUS - DO NOT USE IN PRODUCTION. TESTING PURPOSES ONLY.
def updateField(table, find_field, find_value, change_field, change_value):
    entry = table + ".query.filter_by(" + find_field + "=str(" + find_value + ")).first()"
    entry = eval(entry)
    entry.change_field = change_value
    add(entry)
    return entry


# -------------------- DELETE ------------------------- #


def drop_all():
    db.session.remove()
    db.drop_all()


def remove(entry):
    db.session.delete(entry)
    db.session.commit()


def removeUser(user):
    if isinstance(user, int):
        entry = User.query.get_or_404(user)
    elif isinstance(user, str):
        entry = User.query.filter_by(username=user).first_or_404()
    remove(entry)


def removeClient(client_id):
    if isinstance(client_id, int):
        entry = OAuth2Client.query.get_or_404(client_id)
    else:
        entry = OAuth2Client.query.filter_by(client_id=client_id).first_or_404()
    remove(entry)


def removeAll(table):  # To clear token data
    for each in table.query.all():
        db.session.delete(each)
        remove(each)
